{-# LANGUAGE GADTs, ScopedTypeVariables, FlexibleContexts, RankNTypes, TypeFamilies #-}
{-# OPTIONS_GHC -Wall -fno-warn-name-shadowing #-}
module LanguageGUI where

import Data.IORef
import Graphics.UI.Gtk hiding (Language)
import Graphics.UI.Gtk.Glade
import Data.List
import Language
import Control.Monad
import Data.Map (Map)
import qualified Data.Map as Map

class (Language l) => LanguageGUI l where
  
  render :: Cursor l x a -> String

  renderContextInfo :: Cursor l x a -> String
  
initLanguageGUI :: (LanguageGUI l) => Cursor l l l -> IO ()
initLanguageGUI cursor = do
  initGUI

  Just xml <- xmlNew "language.glade"

  cursorTextView <- xmlGetWidget xml castToTextView "cursor"
  cursorBuffer <- textViewGetBuffer cursorTextView

  inContextTextView <- xmlGetWidget xml castToTextView "inContext"
  inContextBuffer <- textViewGetBuffer inContextTextView

  let guiState = GS {
    cursorBuffer = cursorBuffer,
    inContextBuffer = inContextBuffer,
    cursorH = CH cursor Map.empty
  }

  ref <- newIORef guiState

  cursorTextView `onKeyPress` (cursorKeyPress ref)
  refreshAll ref

  window <- xmlGetWidget xml castToWindow "mainWindow"
  onDestroy window mainQuit
  widgetShowAll window
  
  mainGUI
  

cursorKeyPress :: forall l . (Language l) => IORef (GuiState l) -> Event -> IO Bool
cursorKeyPress ref (Key { eventKeyChar = Just char })
  | char == 'h' = moveCursor [gml, gmu]
  | char == 'j' = moveCursor [gmd, gmr]
  | char == 'k' = moveCursor [gmu, gml]
  | char == 'l' = moveCursor [gmr, gmd]
  | char `elem` loadBookmark = do
      let (Just bid) = findIndex (== char) loadBookmark
      gs@GS { cursorH = CH theCursor @ Cursor {} bmap } <- readIORef ref
      maybe (return True)
            (\(ExistsR bmRoute) -> do
              let (Just cursor') = followRoute theCursor bmRoute
              let bmap' = Map.map (\(ExistsR r) -> ExistsR $ (Language.log cursor') `appendRoute` r) bmap
              writeIORef ref (gs { cursorH = CH (resetLog cursor') bmap' })
              refreshAll ref
              return True
            )
            (Map.lookup bid bmap)

  | char `elem` saveBookmark = do
      let (Just bid) = findIndex (== char) saveBookmark
      gs@GS { cursorH = CH theCursor @ Cursor {} bmap } <- readIORef ref
      let bmap' = Map.insert bid (ExistsR (Route Stop Stop)) bmap
      writeIORef ref (gs { cursorH = CH theCursor bmap' })
      return True
  where
    loadBookmark = "123456789"
    saveBookmark = "!\"£$%^&*("
    gmd = fmap (\(CWM c _) -> ExistsR c) . genericMoveDown
    gmu = fmap (\(CWM c _) -> ExistsR c) . genericMoveUp
    gml = genericMoveLeft
    gmr = genericMoveRight

    moveCursor :: (forall x a . [Cursor l x a -> Maybe (ExistsR l (Cursor l x))]) -> IO Bool
    moveCursor movs = do
      gs@GS { cursorH = CH theCursor @ Cursor {} bmap} <- readIORef ref

      maybe (return True) 
            (\(ExistsR cursor') -> do
              let bmap' = Map.map (\(ExistsR r) -> ExistsR $ (Language.log cursor') `appendRoute` r) bmap
              writeIORef ref (gs { cursorH = CH (resetLog cursor') bmap' })
              refreshAll ref
              return True
            )
            (msum $ map (\f -> f theCursor) movs)
cursorKeyPress _ _ = return False

refreshAll :: IORef (GuiState l) -> IO ()
refreshAll ref = do
  gs@GS { cursorH = CH cursor@Cursor {} _ } <- readIORef ref

  let cursorText = render cursor
  let contextInfo = renderContextInfo cursor

  (cursorBuffer gs) `textBufferSetText` cursorText
  (inContextBuffer gs) `textBufferSetText` contextInfo

data GuiState l = GS {
  cursorBuffer :: TextBuffer,
  inContextBuffer :: TextBuffer,
  cursorH :: CursorHolder l
}
    
  
data CursorHolder l where
  CH :: (LanguageGUI l) => Cursor l a a -> Map Int (ExistsR l (Route l a)) -> CursorHolder l
